"""
Contains typing classes.

NOTE: this module is not intended to be imported at runtime.

"""

import loggings

loggings.warning("this module is not intended to be imported at runtime")

CalendarDict = dict[int, dict[int, list[int]]]



"""
Defines calendar: TradingCalendar, etc.

NOTE: this module is private. All functions and objects are available in the main
`tradingdate` namespace - use that instead.

"""

import datetime
import sys
from typing import TYPE_CHECKING, Iterator, Self

from .calendar_engine import CalendarEngine
from .date import TradingDate, raise_unexpexted_type, split_date

if TYPE_CHECKING:
    from ._typing import CalendarDict

__all__ = [
    "TradingCalendar",
    "YearCalendar",
    "MonthCalendar",
    "WeekCalendar",
    "DayCalendar",
    "NotOnCalendarError",
    "OutOfCalendarError",
]


class TradingCalendar:
    """
    Stores a trading calendar.

    Parameters
    ----------
    cache : CalendarDict
        Calendar dict formatted by `{yyyy: {mm: [dd, ...]}}`, with values
        sorted. Empty lists are not allowed.

    """

    __slots__ = ["id", "cache"]

    def __init__(self, calendar_id: str, cache: "CalendarDict", /) -> None:
        if not cache:
            raise ValueError("empty calendar")
        self.id = calendar_id
        self.cache = cache

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.start} ~ {self.end}, {self.id!r})"

    def __contains__(
        self, value: "int | str | TradingDate | TradingCalendar", /
    ) -> bool:
        if isinstance(value, (int, str, TradingDate)):
            y, m, d = split_date(value)
            return y in self.cache and m in self.cache[y] and d in self.cache[y][m]
        if isinstance(value, TradingCalendar):
            return value.id == self.id
        return False

    def __iter__(self) -> Iterator["TradingDate"]:
        return (
            TradingDate(y, m, d, calendar=self.origin())
            for y in self.cache
            for m in self.cache[y]
            for d in self.cache[y][m]
        )

    def __hash__(self) -> int:
        return hash(str(self))

    def __valur2str(self, value: "int | str | TradingDate | TradingCalendar", /) -> str:
        if value.__class__ is TradingCalendar or self.__class__ is TradingCalendar:
            raise_unsupported_operator(self, value, 2)
        if isinstance(value, int):
            value = str(value)
        elif isinstance(value, (TradingCalendar, TradingDate)):
            value = str(hash(value))
        elif not isinstance(value, str):
            raise_unsupported_operator(self, value, 2)
        return value

    def __eq__(self, value: "int | str | TradingDate | TradingCalendar", /) -> bool:
        if value.__class__ is TradingCalendar and self.__class__ is TradingCalendar:
            return self.id == value.id
        return str(hash(self)) == self.__valur2str(value)

    def __gt__(self, value: "int | str | TradingDate | TradingCalendar", /) -> bool:
        return str(hash(self)) > self.__valur2str(value)

    def __lt__(self, value: "int | str | TradingDate | TradingCalendar", /) -> bool:
        return str(hash(self)) < self.__valur2str(value)

    def __ge__(self, value: "int | str | TradingDate | TradingCalendar", /) -> bool:
        return str(hash(self)) >= self.__valur2str(value)

    def __le__(self, value: "int | str | TradingDate | TradingCalendar", /) -> bool:
        return str(hash(self)) <= self.__valur2str(value)

    @property
    def start(self) -> "TradingDate":
        """Return the starting date of the calendar."""
        y = min(self.cache)
        m = min(self.cache[y])
        d = self.cache[y][m][0]
        return TradingDate(y, m, d, calendar=self.origin())

    @property
    def end(self) -> "TradingDate":
        """Return the ending date of the calendar."""
        y = max(self.cache)
        m = max(self.cache[y])
        d = self.cache[y][m][-1]
        return TradingDate(y, m, d, calendar=self.origin())

    def get_nearest_date_after(self, date: int | str) -> "TradingDate":
        """Get the nearest date after the date (including itself)."""
        y, m, d = split_date(date)
        if y in self.cache:
            ydict = self.cache[y]
            if m in ydict:
                mlist = ydict[m]
                if d in mlist:
                    return TradingDate(y, m, d, calendar=self.origin())
                if d <= mlist[-1]:
                    for dd in mlist:
                        if dd >= d:
                            return TradingDate(y, m, dd, calendar=self.origin())
                    raise RuntimeError("unexpected runtime behavior")
            if m >= 12:
                return self.get_nearest_date_after(f"{y + 1}0101")
            return self.get_nearest_date_after(f"{y}{m + 1:02}01")
        if y < max(self.cache):
            return self.get_nearest_date_after(f"{y + 1}0101")
        raise OutOfCalendarError(
            f"date {date} is out of range [{self.start}, {self.end}]"
        )

    def get_nearest_date_before(self, date: int | str) -> "TradingDate":
        """Get the nearest date before the date (including itself)."""
        y, m, d = split_date(date)
        if y in self.cache:
            ydict = self.cache[y]
            if m in ydict:
                mlist = ydict[m]
                if d in mlist:
                    return TradingDate(y, m, d, calendar=self.origin())
                if d >= mlist[0]:
                    for dd in reversed(mlist):
                        if dd <= d:
                            return TradingDate(y, m, dd, calendar=self.origin())
                    raise RuntimeError("unexpected runtime behavior")
            if m <= 1:
                return self.get_nearest_date_before(f"{y - 1}1231")
            return self.get_nearest_date_before(f"{y}{m - 1:02}31")
        if y > min(self.cache):
            return self.get_nearest_date_before(f"{y - 1}1231")
        raise OutOfCalendarError(
            f"date {date} is out of range [{self.start}, {self.end}]"
        )

    def get_year(self, year: int | str) -> "YearCalendar":
        """Returns a year calendar."""
        y = int(year)
        return YearCalendar(self.id, {y: self.origin().cache[y]})

    def get_month(self, year: int | str, month: int | str) -> "MonthCalendar":
        """Returns a month calendar."""
        y, m = int(year), int(month)
        return MonthCalendar(self.id, {y: {m: self.origin().cache[y][m]}})

    def get_day(
        self, year: int | str, month: int | str, day: int | str
    ) -> "DayCalendar":
        """Returns a day calendar."""
        y, m, d = int(year), int(month), int(day)
        return DayCalendar(self.id, {y: {m: [d]}})

    def get_week(
        self, year: int | str, month: int | str, day: int | str
    ) -> "WeekCalendar":
        """Returns a week calendar."""
        y, m, d = int(year), int(month), int(day)
        w = datetime.date(y, m, d).weekday()
        cdict: "CalendarDict" = {}
        cal = self.origin()
        for date in [
            datetime.date(y, m, d) + datetime.timedelta(days=x)
            for x in range(-w, 7 - w)
        ]:
            y, m, d = date.year, date.month, date.day
            if f"{y}{m:02}{d:02}" in cal:
                if y not in cdict:
                    cdict[y] = {}
                if m not in cdict[y]:
                    cdict[y][m] = [d]
                else:
                    cdict[y][m].append(d)
        return WeekCalendar(self.id, cdict)

    def origin(self) -> Self:
        """Return the original calendar."""
        if self.__class__ is TradingCalendar:
            return self
        return CalendarEngine().get_calendar(self.id)


class YearCalendar(TradingCalendar):
    """Trading year."""

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.asint()}, {self.id!r})"

    def __str__(self) -> str:
        return self.asstr()

    def __int__(self) -> int:
        return self.asint()

    def __hash__(self) -> int:
        return self.asint()

    def __add__(self, value: int, /) -> Self:
        if not isinstance(value, int):
            raise_unexpexted_type(int, value)
        if value < 0:
            return self - abs(value)
        cal = self.origin()
        year_list = list(cal.cache)
        idx = year_list.index(self.asint())
        if idx + value < len(year_list):
            y = year_list[idx + value]
            return cal.get_year(y)
        raise OutOfCalendarError(
            f"year {self} + {value} is out of range [{cal.start}, {cal.end}]"
        )

    def __sub__(self, value: int, /) -> Self:
        if not isinstance(value, int):
            raise_unexpexted_type(int, value)
        if value < 0:
            return self + abs(value)
        cal = self.origin()
        year_list = list(cal.cache)
        idx = year_list.index(self.asint())
        if idx >= value:
            y = year_list[idx - value]
            return cal.get_year(y)
        raise OutOfCalendarError(
            f"year {self} - {value} is out of range [{cal.start}, {cal.end}]"
        )

    def __contains__(
        self, value: "int | str | TradingDate | TradingCalendar", /
    ) -> bool:
        if isinstance(value, (int, str, TradingDate)):
            return super().__contains__(value)
        if isinstance(value, TradingCalendar):
            if not value.id == self.id:
                return False
            if isinstance(value, YearCalendar):
                return False
            return str(hash(self)) in str(hash(value))
        return False

    def next(self) -> Self:
        """Return the next year."""
        return self.end.next().year

    def last(self) -> Self:
        """Return the last year."""
        return self.start.last().year

    def asint(self) -> int:
        """
        Return the year as an integer number equals to `yyyy`.

        Returns
        -------
        int
            An integer representing the year.

        """
        return list(self.cache)[0]

    def asstr(self) -> str:
        """
        Return the year as a string formatted by `yyyy`.

        Returns
        -------
        str
            A string representing the year.

        """
        return str(self.asint())


class MonthCalendar(TradingCalendar):
    """Trading month."""

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({hash(self)}, {self.id!r})"

    def __str__(self) -> str:
        return self.asstr()

    def __int__(self) -> int:
        return self.asint()

    def __hash__(self) -> int:
        y = list(self.cache)[0]
        return int(f"{y}{self.asint():02}")

    def __add__(self, value: int, /) -> Self:
        if not isinstance(value, int):
            raise_unexpexted_type(int, value)
        if value < 0:
            return self - abs(value)
        cal = self.origin()
        month_list = [int(f"{y}{m:02}") for y, x in cal.cache.items() for m in x]
        idx = month_list.index(hash(self))
        if idx + value < len(month_list):
            y = month_list[idx + value]
            return cal.get_year(str(y)[:4]).get_month(str(y)[4:])
        raise OutOfCalendarError(
            f"month {hash(self)} + {value} is out of range [{cal.start}, {cal.end}]"
        )

    def __sub__(self, value: int, /) -> Self:
        if not isinstance(value, int):
            raise_unexpexted_type(int, value)
        if value < 0:
            return self + abs(value)
        cal = self.origin()
        month_list = [int(f"{y}{m:02}") for y, x in cal.cache.items() for m in x]
        idx = month_list.index(hash(self))
        if idx >= value:
            y = month_list[idx - value]
            return cal.get_year(str(y)[:4]).get_month(str(y)[4:])
        raise OutOfCalendarError(
            f"month {hash(self)} - {value} is out of range [{cal.start}, {cal.end}]"
        )

    def __contains__(
        self, value: "int | str | TradingDate | TradingCalendar", /
    ) -> bool:
        if isinstance(value, (int, str, TradingDate)):
            return super().__contains__(value)
        if isinstance(value, TradingCalendar):
            if not value.id == self.id:
                return False
            if isinstance(value, (YearCalendar, MonthCalendar)):
                return False
            return str(hash(self)) in str(hash(value))
        return False

    def next(self) -> Self:
        """Return the next month."""
        return self.end.next().month

    def last(self) -> Self:
        """Return the last month."""
        return self.start.last().month

    def asint(self) -> int:
        """
        Return an integer number equals to `mm`.

        Returns
        -------
        int
            An integer representing the month.

        """
        return list(list(self.cache.values())[0])[0]

    def asstr(self) -> str:
        """
        Return a string formatted by `mm`.

        Returns
        -------
        str
            A string representing the month.

        """
        return f"{self.asint():02}"


class WeekCalendar(TradingCalendar):
    """Trading week."""

    def __repr__(self) -> str:
        y = list(self.cache)[0]
        return f"{self.__class__.__name__}({y} week{self.asstr()}, {self.id!r})"

    def __str__(self) -> str:
        return self.asstr()

    def __int__(self) -> int:
        return self.asint()

    def __hash__(self) -> int:
        return int(f"{hash(self.start) - 1}{self.asstr()}")

    def __add__(self, value: int, /) -> Self:
        if not isinstance(value, int):
            raise_unexpexted_type(int, value)
        if value < 0:
            return self - abs(value)
        week = self
        for _ in range(value):
            week = week.next()
        return week

    def __sub__(self, value: int, /) -> Self:
        if not isinstance(value, int):
            raise_unexpexted_type(int, value)
        if value < 0:
            return self + abs(value)
        week = self
        for _ in range(value):
            week = week.last()
        return week

    def __contains__(
        self, value: "int | str | TradingDate | TradingCalendar", /
    ) -> bool:
        if isinstance(value, (int, str, TradingDate)):
            return super().__contains__(value)
        if isinstance(value, TradingCalendar):
            if not value.id == self.id:
                return False
            if isinstance(value, (YearCalendar, MonthCalendar, WeekCalendar)):
                return False
            return super().__contains__(value.start)
        return False

    def next(self) -> Self:
        """Return the next week."""
        return self.end.next().week

    def last(self) -> Self:
        """Return the last week."""
        return self.start.last().week

    def asint(self) -> int:
        """
        Return an integer number equals to `ww`.

        Returns
        -------
        int
            An integer representing the day.

        """
        return int(self.asstr())

    def asstr(self) -> str:
        """
        Return a string formatted by `ww`.

        Returns
        -------
        str
            A string representing the day.

        """
        return datetime.date(*split_date(self.start)).strftime("%W")


class DayCalendar(TradingCalendar):
    """Trading day."""

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({hash(self)}, {self.id!r})"

    def __str__(self) -> str:
        return self.asstr()

    def __int__(self) -> int:
        return self.asint()

    def __hash__(self) -> int:
        y = list(self.cache)[0]
        m = list(list(self.cache.values())[0])[0]
        return int(f"{y}{m:02}{self.asint():02}")

    def __add__(self, value: int, /) -> Self:
        return (self.end + value).day

    def __sub__(self, value: int, /) -> Self:
        return (self.start - value).day

    def __contains__(
        self, value: "int | str | TradingDate | TradingCalendar", /
    ) -> bool:
        if isinstance(value, (int, str, TradingDate)):
            return super().__contains__(value)
        return False

    def next(self) -> Self:
        """Return the next day."""
        return self.end.next().day

    def last(self) -> Self:
        """Return the last day."""
        return self.start.last().day

    def asint(self) -> int:
        """
        Return an integer number equals to `dd`.

        Returns
        -------
        int
            An integer representing the day.

        """
        return list(list(self.cache.values())[0].values())[0][0]

    def asstr(self) -> str:
        """
        Return a string formatted by `dd`.

        Returns
        -------
        str
            A string representing the day.

        """
        return f"{self.asint():02}"


def raise_unsupported_operator(obj: object, value: object, stacklevel: int) -> None:
    """Raise TypeError."""
    match sys._getframe(stacklevel).f_code.co_name:  # pylint: disable=protected-access
        case "__eq__":
            op = "=="
        case "__gt__":
            op = ">"
        case "__lt__":
            op = "<"
        case "__ge__":
            op = ">="
        case "__le__":
            op = "<="
        case _ as x:
            op = x
    raise TypeError(
        f"{op!r} not supported between instances of {obj.__class__.__name__!r} "
        f"and {value.__class__.__name__!r}"
    )


class NotOnCalendarError(Exception):
    """Raised when date is not on the calendar."""


class OutOfCalendarError(Exception):
    """Raised when date is out of the calendar."""



"""
Defines date: TradingDate, etc.

NOTE: this module is private. All functions and objects are available in the main
`tradingdate` namespace - use that instead.

"""

from typing import TYPE_CHECKING, Callable, Iterator, Self

if TYPE_CHECKING:
    from .calendar import (
        DayCalendar,
        MonthCalendar,
        TradingCalendar,
        WeekCalendar,
        YearCalendar,
    )

__all__ = ["TradingDate", "split_date", "raise_unexpexted_type"]


class TradingDate:
    """
    Represents a trade date on a specified trading calendar.

    Parameters
    ----------
    year : int
        Year number.
    month : int
        Month number.
    day : int
        Day number.
    calendar : TradingCalendar
        Specifies the trading calendar.

    """

    __slots__ = ["calendar", "__date"]

    def __init__(
        self, year: int, month: int, day: int, /, calendar: "TradingCalendar"
    ) -> None:
        self.__date = (year, month, day)
        self.calendar = calendar

    def __eq__(self, value: "Self | int | str | TradingCalendar", /) -> bool:
        if not isinstance(value, (self.__class__, int, str)):
            return value == self
        return self.asint() == int(value)

    def __gt__(self, value: "Self | int | str | TradingCalendar", /) -> bool:
        if not isinstance(value, (self.__class__, int, str)):
            return value < self
        return self.asint() > int(value)

    def __lt__(self, value: "Self | int | str | TradingCalendar", /) -> bool:
        if not isinstance(value, (self.__class__, int, str)):
            return value > self
        return self.asint() < int(value)

    def __ge__(self, value: "Self | int | str | TradingCalendar", /) -> bool:
        if not isinstance(value, (self.__class__, int, str)):
            return value <= self
        return self.asint() >= int(value)

    def __le__(self, value: "Self | int | str | TradingCalendar", /) -> bool:
        if not isinstance(value, (self.__class__, int, str)):
            return value >= self
        return self.asint() <= int(value)

    def __add__(self, value: int, /) -> Self:
        if not isinstance(value, int):
            raise_unexpexted_type(int, value)
        if value < 0:
            return self - abs(value)
        y, m, d = split_date(self.asstr())
        month = self.calendar.cache[y][m]
        idx = month.index(d)
        if idx + value < len(month):
            d = month[idx + value]
            return self.__class__(y, m, d, calendar=self.calendar)
        value -= len(month) - idx
        return self.calendar.get_nearest_date_after(f"{y}{m + 1:02}01") + value

    def __sub__(self, value: int, /) -> Self:
        if not isinstance(value, int):
            raise_unexpexted_type(int, value)
        if value < 0:
            return self + abs(value)
        y, m, d = split_date(self.asstr())
        month = self.calendar.cache[y][m]
        idx = month.index(d)
        if idx >= value:
            d = month[idx - value]
            return self.__class__(y, m, d, calendar=self.calendar)
        value -= idx + 1
        return self.calendar.get_nearest_date_before(f"{y}{m - 1:02}31") - value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.asstr()})"

    def __str__(self) -> str:
        return self.asstr()

    def __int__(self) -> int:
        return self.asint()

    def __hash__(self) -> int:
        return self.asint()

    def next(self) -> Self:
        """Returns the next date."""
        y, m, d = self.__date
        return self.calendar.get_nearest_date_after(f"{y}{m:02}{d + 1:02}")

    def last(self) -> Self:
        """Returns the last date."""
        y, m, d = self.__date
        return self.calendar.get_nearest_date_before(f"{y}{m:02}{d - 1:02}")

    def iterate_until(
        self,
        stop: "TradingDate | int | str",
        step: int = 1,
        /,
        *,
        include_end: bool = False,
    ) -> "DateRange":
        """
        Returns an iterator of trade dates from `self` (inclusive) to
        `stop` (inclusive or exclusive, determined by argument) by `step`.

        Equivalent to `daterange(self, stop, step)` if `inclusive` is
        False.

        Parameters
        ----------
        end : TradingDate | int | str
            End date.
        step : int, optional
            Step, by default 1.
        include_end : bool, optional
            Determines whether `stop` is inclusive in the iterator.

        Returns
        -------
        DateRange
            Iterator of trade dates.

        """
        return DateRange(self, stop, step, include_end=include_end)

    def asint(self) -> int:
        """
        Return an integer number equals to `yyyymmdd`.

        Returns
        -------
        int
            An integer representing the date.

        """
        return int(self.asstr())

    def asstr(self) -> str:
        """
        Return a string formatted by `yyyymmdd`.

        Returns
        -------
        str
            A string representing the date.

        """
        y, m, d = self.__date
        return f"{y}{m:02}{d:02}"

    @property
    def year(self) -> "YearCalendar":
        """Calendar of the year."""
        return self.calendar.get_year(self.__date[0])

    @property
    def month(self) -> "MonthCalendar":
        """Calendar of the month."""
        y, m, _ = self.__date
        return self.calendar.get_month(y, m)

    @property
    def week(self) -> "WeekCalendar":
        """Calendar of the week."""
        return self.calendar.get_week(*self.__date)

    @property
    def day(self) -> "DayCalendar":
        """Calendar of the day."""
        return self.calendar.get_day(*self.__date)


class DateRange:
    """
    Returns an iterator of trade dates from `start` (inclusive) to
    `stop` (inclusive or exclusive, determined by argument) by `step`.

    """

    def __init__(
        self,
        start: TradingDate,
        stop: "TradingDate | int | str",
        step: int = 1,
        /,
        *,
        include_end: bool = False,
    ) -> None:
        self.__start = start
        self.__stop = stop
        self.__step = step
        self.__include_end = include_end

    def __repr__(self) -> str:
        rstr = f"{self.__class__.__name__}({self.__start}, {self.__stop}"
        if self.__step != 1:
            rstr += f", {self.__step}"
        if self.__include_end:
            rstr += ", include_end=True"
        rstr += ")"
        return rstr

    def __iter__(self) -> Iterator[TradingDate]:
        if self.__include_end:
            return self.__iter_inclusively()
        return self.__iter_exclusively()

    def __iter_exclusively(self) -> Iterator[TradingDate]:
        date, stop, step = self.__start, self.__stop, self.__step
        if step == 1:
            while date < stop:
                yield date
                date = date.next()
        elif step == -1:
            while date > stop:
                yield date
                date = date.last()
        elif step == 0:
            raise ValueError("step must not be zero")
        elif step > 0:
            while date < stop:
                yield date
                date = date + step
        else:
            while date > stop:
                yield date
                date = date + step

    def __iter_inclusively(self) -> Iterator[TradingDate]:
        date, stop, step = self.__start, self.__stop, self.__step
        if step == 1:
            while date <= stop:
                yield date
                date = date.next()
        elif step == -1:
            while date >= stop:
                yield date
                date = date.last()
        elif step == 0:
            raise ValueError("step must not be zero")
        elif step > 0:
            while date <= stop:
                yield date
                date = date + step
        else:
            while date >= stop:
                yield date
                date = date + step

    def tolist(self) -> list[TradingDate]:
        """Equivalent to `list(self)`."""
        return list(self)

    def find_every_year(self) -> list["YearCalendar"]:
        """
        Return a list of YearCalendar between `start` and `stop` by `step`.

        """
        return sorted(set(x.year for x in self))

    def find_every_month(self) -> list["MonthCalendar"]:
        """
        Return a list of MonthCalendar between `start` and `stop` by `step`.

        """
        return sorted(set(x.month for x in self))

    def find_every_week(self) -> list["WeekCalendar"]:
        """
        Return a list of WeekCalendar between `start` and `stop` by `step`.

        """
        return sorted(set(x.week for x in self))

    def apply[T](self, func: Callable[[TradingDate], T]) -> list[T]:
        """Apply a function on each TradingDate and return a list."""
        return [func(x) for x in self]


def split_date(date: TradingDate | int | str) -> tuple[int, int, int]:
    """Split date to int numbers: year, month, and day."""
    datestr = str(date)
    return int(datestr[:-4]), int(datestr[-4:-2]), int(datestr[-2:])


def raise_unexpexted_type(typ: type, value: object) -> None:
    """Raise TypeError."""
    raise TypeError(f"expected {typ.__name__}, got {type(value).__name__} instead")



"""
Contains the core of tradingdate: get_trading_date(), get_calendar(), etc.

NOTE: this module is private. All functions and objects are available in the main
`tradingdate` namespace - use that instead.

"""

from typing import Iterator, Literal

from .calendar import NotOnCalendarError, TradingCalendar
from .calendar_engine import CalendarEngine
from .date import DateRange, TradingDate

__all__ = ["date", "daterange", "get_calendar", "make_calendar"]


def date(
    date: int | str,
    /,
    calendar_id: str = "chinese",
    missing: Literal["use_next", "use_before", "raise"] = "use_before",
) -> TradingDate:
    """
    Returns a `TradingDate` object.

    Parameters
    ----------
    date : int | str
        The date.
    calendar_id : str, optional
        Calendar id, by default "chinese".
    missing : Literal["use_next", "use_before", "raise"], optional
        Used when `date` is not found in the calendar. If "use_next",
        return the nearest trade date after `date`; if "use_before",
        return the nearest trade date before it; if "raise", raise
        error. By default "use_before".

    Returns
    -------
    TradingDate
        Trade date.

    """

    calendar = get_calendar(calendar_id)
    match missing:
        case "use_next":
            return calendar.get_nearest_date_after(date)
        case "use_before":
            return calendar.get_nearest_date_before(date)
        case "raise":
            raise NotOnCalendarError(f"date {date} is not on the calendar")
        case _ as x:
            raise ValueError(f"invalid value for argument 'not_exist': {x!r}")


def daterange(
    start: int | str | None = None,
    end: int | str | None = None,
    step: int = 1,
    /,
    calendar_id: str = "chinese",
    include_end: bool = False,
) -> DateRange:
    """
    Returns an iterator of trade dates from `start` (inclusive) to
    `stop` (exclusive) by `step`.

    Parameters
    ----------
    start : int | str | None, optional
        Start date, by default None.
    end : int | str | None, optional
        End date, by default None.
    step : int, optional
        Step, by default 1.
    calendar_id : str, optional
        Calendar id, by default "chinese".
    include_end : bool, optional
        Whether the end date should be included, by default False.

    Returns
    -------
    DateRange
        Iterator of trade dates.

    """
    calendar = get_calendar(calendar_id)
    date = calendar.start if start is None else calendar.get_nearest_date_after(start)
    end = calendar.end if end is None else end
    return DateRange(date, end, step, include_end=include_end)


def get_calendar(calendar_id: str = "chinese") -> TradingCalendar:
    """
    Returns a `TradingCalendar` object.

    Parameters
    ----------
    calendar_id : str, optional
        Calendar id, by default "chinese".

    Returns
    -------
    TradingCalendar
        Calendar.

    """
    engine = CalendarEngine()
    match calendar_id:
        case "chinese":
            cal = engine.get_chinese_calendar(TradingCalendar)
        case _ as x:
            cal = engine.get_calendar(x)
    return cal


def make_calendar(
    calendar_id: str, dates: Iterator[int | str | TradingDate]
) -> TradingCalendar:
    """
    Make a new calendar and register it in the engine.

    Parameters
    ----------
    calendar_id : str
        Calendar id.
    dates : Iterator[int | str | TradingDate]
        Dates formatted by `yyyymmdd`.

    Returns
    -------
    TradingCalendar
        Calendar.

    """
    engine = CalendarEngine()
    engine.register_calendar(TradingCalendar, calendar_id, list(dates))
    return engine.get_calendar(calendar_id)



"""
Provides the tool for getting calendars: CalendarEngine.

NOTE: this module is private. All functions and objects are available in the main
`tradingdate` namespace - use that instead.

"""

import datetime
from typing import TYPE_CHECKING

import chinese_calendar

if TYPE_CHECKING:
    from ._typing import CalendarDict
    from .calendar import TradingCalendar
    from .date import TradingDate

__all__ = ["CalendarEngine"]


class CalendarEngine:
    """
    Calendar engine.

    Output should be dicts formatted by `{yyyy: {mm: [dd, ...]}}`. The
    numbers are sorted.

    """

    __calendar_cache: dict[str, "TradingCalendar"] = {}

    def get_chinese_calendar(
        self, constructor: type["TradingCalendar"]
    ) -> "TradingCalendar":
        """Get the chinese calendar."""
        if "chinese" not in self.__calendar_cache:
            y, m, d = 2004, 1, 1
            cal: "CalendarDict" = {y: {m: []}}
            try:
                workdays = chinese_calendar.get_workdays(
                    datetime.date(y, m, d),
                    datetime.date(datetime.datetime.now().year, 12, 31),
                )
            except NotImplementedError as e:
                e.add_note(
                    "please try 'pip install --upgrade chinesecalendar' in the console"
                )
                raise e
            for x in workdays:
                if x.year == y:
                    if x.month == m:
                        d = x.day
                        cal[y][m].append(d)
                    else:
                        m, d = x.month, x.day
                        cal[y][m] = [d]
                else:
                    y, m, d = x.year, x.month, x.day
                    cal[y] = {}
                    cal[y][m] = [d]
            self.__calendar_cache["chinese"] = constructor("chinese", cal)
        return self.__calendar_cache["chinese"]

    def register_calendar(
        self,
        constructor: type["TradingCalendar"],
        calendar_id: str,
        date_list: list["int | str | TradingDate"],
    ) -> None:
        """Register a calendar."""
        if calendar_id in self.__calendar_cache:
            raise ValueError(f"calendar_id already exists: {calendar_id!r}")
        datestr_list: list[str] = sorted({str(x) for x in date_list})
        y, m, d = -1, -1, -1
        cal: "CalendarDict" = {}
        for datestr in datestr_list:
            if len(datestr) <= 4:
                raise ValueError(f"invalid date: {datestr}")
            yy, mm, dd = int(datestr[:-4]), int(datestr[-4:-2]), int(datestr[-2:])
            if yy == y:
                if mm == m:
                    self.__check_day(d := dd)
                    cal[y][m].append(d)
                else:
                    self.__check_month(m := mm)
                    self.__check_day(d := dd)
                    cal[y][m] = [d]
            else:
                self.__check_year(y := yy)
                self.__check_month(m := mm)
                self.__check_day(d := dd)
                cal[y] = {}
                cal[y][m] = [d]
        self.__calendar_cache[calendar_id] = constructor(calendar_id, cal)

    def get_calendar(self, calendar_id: str) -> "TradingCalendar":
        """Get a calendar."""
        return self.__calendar_cache[calendar_id]

    def __check_year(self, year: int) -> None:
        if year < 0:
            raise ValueError(f"invalid year number: {year}")

    def __check_month(self, month: int) -> None:
        if not 1 <= month <= 12:
            raise ValueError(f"invalid month number: {month}")

    def __check_day(self, day: int) -> None:
        if not 1 <= day <= 31:
            raise ValueError(f"invalid day number: {day}")



"""
# tradingdate
Manages trade dates.

## Usage
### Get a calendar
```py
>>> import tradingdate as td
>>> cal = td.get_calendar("chinese") # by default "chinese"
>>> cal
TradingCalendar(20040102 ~ 20261231, 'chinese')
>>> list(cal)[:3]
[TradingDate(20040102), TradingDate(20040105), TradingDate(20040106)]
```

### Get a trading date
```py
>>> date = td.date(20250116)
>>> date
TradingDate(20250116)
>>> print(date.year, date.month, date.day)
2025 01 16

>>> date - 20
TradingDate(20241218)
>>> date + 100
TradingDate(20250617)

>>> date.week.start, date.week.end
(TradingDate(20250113), TradingDate(20250117))
>>> date.month.start, date.month.end
(TradingDate(20250102), TradingDate(20250127))
>>> date.year.start, date.year.end
(TradingDate(20250102), TradingDate(20251231))
```

### Get iterator of trading dates
```py
>>> list(td.daterange(20250101, 20250107))
[TradingDate(20250102), TradingDate(20250103), TradingDate(20250106)]
```

### Make a new calendar
```py
>>> td.make_calendar("user-defined", [20250101, 20250115, 20250201])
TradingCalendar(20250101 ~ 20250201, 'user-defined')
>>> list(td.daterange(20250101, 20250131, calendar_id="user-defined"))
[TradingDate(20250101), TradingDate(20250115)]
```

## See Also
### Github repository
* https://github.com/Chitaoji/tradingdate/

### PyPI project
* https://pypi.org/project/tradingdate/

## License
This project falls under the BSD 3-Clause License.

"""

from typing import List

from . import core
from ._version import __version__
from .core import *

__all__: List[str] = []
__all__.extend(core.__all__)
